#ifndef __ConQueue__
#define __ConQueue__
#include <queue>
#include <mutex>
#include <condition_variable>
#include <set>
#include <string>
#include "Logger.hpp"

class Comparator
{
public:
    inline bool operator()(const struct ibv_wc *c1, const struct ibv_wc *c2)
        const
    {
        struct MessageHeader *req1 = (struct MessageHeader *)c1->wr_id;
        struct MessageHeader *req2 = (struct MessageHeader *)c2->wr_id;
        if (req1->keySize != req2->keySize)
            return true;
        char *key1 = (char *)req1 + MessageHeaderSize;
        char *key2 = (char *)req2 + MessageHeaderSize;
        for (uint32_t i = 0; i < req1->keySize; i++)
        {
            if (key1[i] != key2[i])
                return true;
        }
        return false;
    }
};

class ConcurrentQueue
{
private:
    std::queue<struct ibv_wc *> queue1;
    std::queue<struct ibv_wc *> queue2;
    std::mutex queueMutex;
    std::set<struct ibv_wc *, Comparator> runningRequests;
    std::condition_variable queueCv;

public:
    void push(struct ibv_wc *const &data);
    bool empty();
    struct ibv_wc *try_pop();
    void removeFromSet(struct ibv_wc *data);
    void wait_and_pop(struct ibv_wc *&popped_value);
};
#endif

/*

template <typename Data>
class ConcurrentQueue
{
private:
    std::queue1<Data> queue;
    std::queue2<Data> queue;
    std::mutex queueMutex;
    std::condition_variable queueCv;

public:
    void push(Data const &data)
    {
        std::unique_lock<std::mutex> lock(queueMutex);
        queue1.push(data);
        lock.unlock();
        queueCv.notify_one();
    }

    bool empty()
    {
        std::unique_lock<std::mutex> lock(queueMutex);
        return queue.empty();
    }

    bool try_pop(Data &popped_value)
    {
        std::unique_lock<std::mutex> lock(queueMutex);
        if (queue2.empty())
        {
            return false;
        }

        popped_value = queue.front();
        queue.pop();
        return true;
    }

    void wait_and_pop(Data &popped_value)
    {
        std::unique_lock<std::mutex> lock(queueMutex);
        queueCv.wait(lock, [&]{return queue.size() > 0;});
        popped_value = queue.front();
        queue.pop();
    }
};
#endif
*/