#ifndef __RDMASERVERENDPOINTGROUP__
#define __RDMASERVERENDPOINTGROUP__

#include <rdma/rdma_cma.h>
#include <rdma/rdma_verbs.h>
#include <stdint.h>
#include <thread>
#include <netdb.h>
#include <arpa/inet.h>
#include <map>
#include <vector>
#include <mutex>
#include <shared_mutex>

#include "RdmaReplicationEndpoint.hpp"
#include "RdmaSalEndpoint.hpp"
#include "RdmaEndpointGroup.hpp"
#include "CqEventData.hpp"
#include "Executor.hpp"
#include "RdmaCmProcessor.hpp"
#include "RdmaSalCqProcessor.hpp"
#include "RdmaRepCqProcessor.hpp"
#include "Logger.hpp"

class RdmaServerEndpointGroup : public RdmaEndpointGroup
{
    /*
     * Variables to maintain Group state
     */
    static int CONN_STATE_INITIALIZED;
    static int CONN_STATE_BINDED;
    static int CONN_STATE_CONNECTED;
    static int CONN_STATE_CLOSED;

    struct rdma_cm_id *_cm_id{NULL};
    RdmaCmProcessor *_cmProcessor{NULL};
    RdmaSalCqProcessor *_salCqProcessor{NULL};
    RdmaRepCqProcessor *_repCqProcessor{NULL};

    Executor *_executor;

    /*
     * variables to maintain queue state
     */
    int _sendQueueSize{0};
    int _recvQueueSize{0};
    int _compQueueSize{0};
    int _sendMsgSize{0};
    int _recvMsgSize{0};
    rocksdb::DB *_db;

    mutable std::shared_mutex _salMutex;
    mutable std::shared_mutex _repMutex;

    void clientClose();

public:
    RdmaServerEndpointGroup(int sendQueueSize, int recvQueueSize, int compQueueSize,
                            int sendMsgSize, int recvMsgSize, std::string dbpath);
    // void setExecutor(Executor *executor);
    void bind(const char *ip, const char *port, int backlog);

    struct ibv_cq *createSalCq(struct rdma_cm_id *id);
    struct ibv_cq *createRepCq(struct rdma_cm_id *id);
    void dispatchRepCqEvents(ibv_wc wc[], int size);
    void dispatchSalCqEvents(ibv_wc wc[], int size);
    void setExecutor(Executor *ex);
    void startCmProcessor(bool newThread);
    void processCmEvent(struct rdma_cm_event *event);
    void createEpCmEvent(struct rdma_cm_event *event);
    void close();
};

#endif