# How to run server
> To edit run time configuration like db path and server parameters edit prop.config file <br>
> Step 1 ensuring RDMA drivers are loaded <br>
> Step 2 Ensuring rocksdb library is loaded <br>
> Step 3 Compile : make server <br>
> Step 4  run make server <br>

# To do list
>### High Priority:
>- [ ] consensus
>- [ ] failure detection and recovery
>- [ ] explore multicast
>- [x] Added logging api
>- [x] add multi connection in client
>- [x] add yscb 
>- [ ] add dynamic hashing in client code
>- [ ] add waiting for response form followers and wait for invalidation


>### Low Priority
>- [ ] Handle RDMA_CM_EVENT_TIMEWAIT_EXIT when closed connection
>- [x] Handle Work Request Flushed Error on qp
>- [ ] Add condition variable in ConcurrentQueue to avoid busy waiting
 
# References:
> [Get CQ Event Rdma Mojo](https://www.rdmamojo.com/2013/03/09/ibv_get_cq_event/) <br>
> [Poll CQ](https://man7.org/linux/man-pages/man3/ibv_poll_cq.3.html) <br>
> [Get CQ Event](https://man7.org/linux/man-pages/man3/ibv_get_cq_event.3.html) <br>
> [Delegating Constructor](https://docs.microsoft.com/en-us/cpp/cpp/delegating-constructors?view=msvc-170) <br>
> [Understanding C++ Compilation](https://www.toptal.com/c-plus-plus/c-plus-plus-understanding-compilation) <br>
> [C++ Logging](https://www.codeproject.com/Tips/987850/Logging-in-Cplusplus) <br>
> [Creating Readme.md](https://www.mygreatlearning.com/blog/readme-file/) <br>

# Git making changes to file untrack/track from index
>To get list of unchanged files
>git ls-files -v|grep '^h'
>git update-index --assume-unchanged FILE_NAME<br>
>git update-index --no-assume-unchanged readme.md


# Example to set cpu affinity:
```c++
    cpu_set_t cpuset;
    CPU_ZERO(&cpuset);
    CPU_SET(i, &cpuset);
    int rc = pthread_setaffinity_np(threads[i].native_handle(),
                                    sizeof(cpu_set_t), &cpuset);
    if (rc != 0) {
      std::cerr << "Error calling pthread_setaffinity_np: " << rc << "\n";
    }
```
# Logging API Example
```
   LOG_ALWAYS("<=============================== START OF PROGRAM ===============================>");
   // Log message using Direct Interface
   // Log Level: LOG_ERROR
   LOG_ERROR("Message Logged using Direct Interface, Log level: LOG_ERROR");
   LOG_ALARM("Message Logged using Direct Interface, Log level: LOG_ALARM");
   LOG_ALWAYS("Message Logged using Direct Interface, Log level: LOG_ALWAYS");
   LOG_INFO("Message Logged using Direct Interface, Log level: LOG_INFO");
   LOG_BUFFER("Message Logged using Direct Interface, Log level: LOG_BUFFER");
   LOG_TRACE("Message Logged using Direct Interface, Log level: LOG_TRACE");
   LOG_DEBUG("Message Logged using Direct Interface, Log level: LOG_DEBUG");

   // Log message C++ Interface
   Logger* pLogger = NULL; // Create the object pointer for Logger Class
   pLogger = Logger::getInstance();

   pLogger->error("Message Logged using C++ Interface, Log level: LOG_ERROR");
   
   std::ostringstream ss;
   ss << endl;
   ss << "\t" << "My Contact Details:" << endl;
   
   //pLogger->enableConsoleLogging();
   pLogger->updateLogLevel(LOG_LEVEL_INFO);

   // Log ostringstream ss to all the log levels
   LOG_ALWAYS("Logging ostringstream using Direct Interface");
   LOG_ERROR(ss);

   Logger::getInstance()->buffer("Logging ostringstream using C++ Interface");
   Logger::getInstance()->error(ss);
   
   LOG_ALWAYS("<========= END OF PROGRAM ==========>");
```

