#include <TaskThread.hpp>
void ConcurrentQueue::push(struct ibv_wc *const &data)
{
    std::unique_lock<std::mutex> lock(queueMutex);
    std::cout<<"putting data\n";
    queue1.push(data);
    lock.unlock();
    queueCv.notify_one();
}

bool ConcurrentQueue::empty()
{
    std::unique_lock<std::mutex> lock(queueMutex);
    return queue1.empty() && queue2.empty();
}

struct ibv_wc *ConcurrentQueue::try_pop()
{
    struct ibv_wc *value = NULL;
    std::unique_lock<std::mutex> lock(queueMutex);
    if (queue2.empty())
    {
        queueCv.wait(lock, [&]
                     { return queue1.size() > 0; });
        value = queue1.front();
        queue1.pop();
    }
    else
    {
        value = queue2.front();
        queue2.pop();
    }
    //We only want to handle 
    if (value->opcode != IBV_WC_RECV)
    {
        return value;
    }
    //  std::cout<<"value "<<value<<std::endl;
    if (runningRequests.empty())
    {
        runningRequests.insert(value);
        return value;
    }

    auto it = runningRequests.find(value);
    if (it != runningRequests.end())
    {
        queue2.push(value);
        // std::cout<<"found putting in 2"<<std::endl;
        return NULL;
    }
    return value;
}
void ConcurrentQueue::removeFromSet(struct ibv_wc *data)
{
    std::unique_lock<std::mutex> lock(queueMutex);
    // std::cout<<"removing"<<data<<std::endl;
    runningRequests.erase(data);
}
void ConcurrentQueue::wait_and_pop(struct ibv_wc *&popped_value)
{
    std::unique_lock<std::mutex> lock(queueMutex);
    queueCv.wait(lock, [&]
                 { return queue1.size() > 0; });
    popped_value = queue1.front();
    queue1.pop();
}