#include "RdmaReplicationEndpoint.hpp"

RdmaReplicationEndpoint::RdmaReplicationEndpoint(struct rdma_cm_id *id, struct ibv_cq *completionQueue, int sendQueueSize,
                                                 int recvQueueSize, int sendMsgSize, int recvMsgSize, rocksdb::DB *db)
    : RdmaEndpoint(id, completionQueue, sendQueueSize, recvQueueSize, sendMsgSize, recvMsgSize), _db(db)
{
}

void RdmaReplicationEndpoint::processSendCompletion(struct ibv_wc *data)
{
    //CPPLog::LOG_INFO("send completion");
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    _sendBuffers.push((char *)data->wr_id);
}
void RdmaReplicationEndpoint::processRecvCompletion(struct ibv_wc *data)
{
    //CPPLog::LOG_INFO("recv completion");
    std::cout << "Replication recieve" << (char *)(data->wr_id) << "\n";
    char *request = new char[data->byte_len];
    memcpy(request, (void *)data->wr_id, data->byte_len);
    rdma_post_recv(_cm_id, (void *)data->wr_id, (void *)data->wr_id, _recvMsgSize, _recvMr);
    delete[] request;
}

int RdmaReplicationEndpoint::sendMessage(const char *buffer, uint32_t size)
{
    if (size > (uint32_t)_sendMsgSize)
        return -1;
    void *sendBuffer = nullptr;
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    if (_sendBuffers.size() == 0)
        return -1;
    sendBuffer = _sendBuffers.front();
    _sendBuffers.pop();
    lock.unlock();
    memcpy(sendBuffer, buffer, size);
    return rdma_post_send(_cm_id, sendBuffer, sendBuffer, size, _sendMr, 0);
}
