#include "RdmaSalCqProcessor.hpp"
RdmaSalCqProcessor::RdmaSalCqProcessor(Executor *ex, ibv_context *verbs, int compQueueSize)
    : _executor(ex)
{
    _compChannel = ibv_create_comp_channel(verbs);
    if (_compChannel == NULL)
    {
        CPPLog::LOG_ERROR("SalCqProcessr : ibv_create_comp_channel failed");
        return;
    }
    _completionQueue = ibv_create_cq(verbs, compQueueSize, NULL, _compChannel, 0);
    if (_completionQueue == NULL)
    {
        CPPLog::LOG_INFO("SalCqProcessr : ibv_create_cq failed");
        return;
    }
    int ret = ibv_req_notify_cq(_completionQueue, 0);
    if (ret)
    {
        CPPLog::LOG_INFO("SalCqProcessr : ibv_req_notify_cq failed");
    }
}
struct ibv_cq *RdmaSalCqProcessor::getCq()
{
    return _completionQueue;
}
void RdmaSalCqProcessor::start()
{
    CPPLog::LOG_ALWAYS("SalCqProcessr : starting process CQ events");
    _compQueueThread = new std::thread(&RdmaSalCqProcessor::processCQEvents, this);
    pthread_setname_np(_compQueueThread->native_handle(),"SalCQProcessor");
}
void RdmaSalCqProcessor::processCQEvents()
{
    int ret = 0;
    struct ibv_cq *cq;
    void *context;
    const int nevent = 10;
    struct ibv_wc wc_array[nevent];
    while (!_stop)
    {
        /*
         * get_CQ_event is a blocking call and it wait save some cpu cycles but.
         * it might not be that efficient compared to polling
         */
        ret = ibv_get_cq_event(_compChannel, &cq, &context);
        if (ret == -1)
        {
            CPPLog::LOG_ERROR("SalCqProcessr : ibv_get_cq_event failed");
            close();
        }
        ibv_ack_cq_events(cq, 1);
        /*
         * Create a request for next completion cycle
         */
        ret = ibv_req_notify_cq(_completionQueue, 0);
        if (ret)
        {
            CPPLog::LOG_ERROR("SalCqProcessr : ibv_req_notify_cq failed");
            close();
        }
        ret = ibv_poll_cq(cq, nevent, wc_array);
        if (ret < 0)
        {
            CPPLog::LOG_ERROR("SalCqProcessr : ibv_poll_cq failed");
            close();
        }
        if (ret == 0)
            continue;
        for (int i = 0; i < ret; i++)
        {
            if (wc_array[i].status != IBV_WC_SUCCESS)
            {
                std::ostringstream ss;
                ss<< "RdmaSalCqProcessor : failed work completion : ";
                ss<<ibv_wc_status_str(wc_array[i].status)<<"on qp"<<wc_array[i].qp_num;
                CPPLog::LOG_ERROR(ss);
                continue;
            }
            struct ibv_wc *data = new struct ibv_wc(wc_array[i]);
            /*
             * vendor_err is set to check whether the request came from sal or followers
             * data->vendor_err = 0;
             */
            _executor->submit(data);
        }
    }
}

void RdmaSalCqProcessor::close()
{
    _stop = true;
    if (_compQueueThread != NULL)
        _compQueueThread->join();
    delete _compQueueThread;
}