package hpdos_rdma_offloaded.handler;

import java.net.SocketAddress;
import java.net.InetSocketAddress;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import com.google.common.base.Stopwatch;
import com.ibm.darpc.DaRPCClientEndpoint;
import com.ibm.darpc.DaRPCClientGroup;
import com.ibm.darpc.DaRPCFuture;
import com.ibm.darpc.DaRPCStream;
import com.ibm.disni.util.StopWatch;

import org.rocksdb.RocksDBException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import hpdos_rdma_offloaded.lib.Follower;
import hpdos_rdma_offloaded.lib.Packet;
import hpdos_rdma_offloaded.protocol.InvalidationRequest;
import hpdos_rdma_offloaded.protocol.InvalidationResponse;
import hpdos_rdma_offloaded.protocol.InvalidationRpcProtocol;
import hpdos_rdma_offloaded.protocol.Request;
import hpdos_rdma_offloaded.protocol.Response;
import hpdos_rdma_offloaded.protocol.RpcProtocol;

public class NetworkHandler {
    final static Logger logger = LoggerFactory.getLogger("NetworkHandler");
    private ExecutorService executorService;
    // public HashMap<String, DaRPCStream<Request, Response>> streams;
    public static ConcurrentHashMap<String, DaRPCStream<InvalidationRequest, InvalidationResponse>> invalidationStreams =  new ConcurrentHashMap<>();
    public static List<Follower> followers = new ArrayList<>();
    boolean isMaster;
    // ConcurrentHashMap<SocketAddress,DaRPCStream<InvalidationRequest, InvalidationResponse>> invalidationClientEpStreams;
    public NetworkHandler(boolean isMaster){
        this.executorService = Executors.newFixedThreadPool(10);
        this.isMaster = isMaster;
    }

    public NetworkHandler(){}

    public void setClientStreams(ConcurrentHashMap<SocketAddress,DaRPCStream<InvalidationRequest, InvalidationResponse>> invalidationClientEpStreams)
    {
        // this.invalidationClientEpStreams = invalidationClientEpStreams;
    }
    
    public void connectToSal(String ip) throws Exception{
        Thread.sleep(3000);
        logger.debug("Making rdma connection to client for invalidation request.");
        // Follower follower = new Follower();
        // follower.setFollowerId(uid);
        // follower.setIpAddress(ip);
        InvalidationRpcProtocol rpcProtocol = new InvalidationRpcProtocol();
        DaRPCClientGroup<InvalidationRequest, InvalidationResponse> group = DaRPCClientGroup.createClientGroup(rpcProtocol, 100, 0, 16, 16); 
        InetSocketAddress address = new InetSocketAddress(ip, 1922);
        DaRPCClientEndpoint<InvalidationRequest, InvalidationResponse> clientEp = group.createEndpoint();
        clientEp.connect(address, 1000);
        DaRPCStream<InvalidationRequest, InvalidationResponse> stream = clientEp.createStream();
        NetworkHandler.invalidationStreams.put(ip, stream);
    }

    public void connectToFollower(String uid , String ip) throws Exception {
        Thread.sleep(3000);
        logger.debug("Making rdma connection to follower.");
        Follower follower = new Follower();
        follower.setFollowerId(uid);
        follower.setIpAddress(ip);
        RpcProtocol rpcProtocol = new RpcProtocol();
        DaRPCClientGroup<Request, Response> group = DaRPCClientGroup.createClientGroup(rpcProtocol, 100, 0, 16, 16); 
        InetSocketAddress address = new InetSocketAddress(ip, 1921);
        DaRPCClientEndpoint<Request, Response> clientEp = group.createEndpoint();
        clientEp.connect(address, 1000);
        DaRPCStream<Request, Response> stream = clientEp.createStream();
        follower.setStream(stream);
        NetworkHandler.followers.add(follower);
    }

    // Change the Futre to CompletableFuture to achieve correct asynchronousness
    public void create(Packet packet) throws InterruptedException, ExecutionException,RocksDBException{
        ReplicationHandler replicationHandler = new ReplicationHandler();
        logger.debug("Received create request for key/value: " + new String(packet.getKey()) + "/" + new String(packet.getValue()));
        Stopwatch stopWatch = Stopwatch.createUnstarted();
        stopWatch.start();
        if (this.isMaster)
        {
            // logger.debug("Starting replication");
            // Future<Response> futureReplication = this.executorService.submit(()->{
            //     Response replicationResponse = new Response();
            //     // Write code to replicate the data to other nics
            //     ReplicationHandler.replicateMetadata(packet);
            //     return replicationResponse;
            
            // replicationHandler.replicateMetadata(packet);
            Set<Callable<DaRPCFuture<Request, Response>>> callables = new HashSet<>();
        for(Follower follower : NetworkHandler.followers){
            callables.add(()->{
                Request request = new Request();
                Response response = new Response();
                request.setKey(packet.getKey());
                request.setValue(packet.getValue());
                request.setRequestType(packet.getOperationType());
                DaRPCFuture<Request, Response> future = follower.getStream().request(request, response, false);
                return future;
            });
        }
        List<Future<DaRPCFuture<Request, Response>>> futures = executorService.invokeAll(callables);
        
        for (Future<DaRPCFuture<Request, Response>> future: futures) {
            Response response = future.get().get(); 
        }
            stopWatch.stop();
        } 
        
        logger.debug("Future time: " + stopWatch.elapsed(TimeUnit.MICROSECONDS));
        stopWatch.reset();
        // stopWatch.start();
        // Response response = futureReplication.get();     
        // stopWatch.stop();
        // logger.debug("Replication time: " + stopWatch.elapsed(TimeUnit.MICROSECONDS));
        logger.debug("Replication complete");
        
    }

    public void read(Packet packet) throws RocksDBException,InterruptedException,ExecutionException{
        logger.debug("Received read request for key/value: " + new String(packet.getKey()));
        if (this.isMaster)
        {    
            Future<Response> futureReplication = this.executorService.submit(()->{
                Response replicationResponse = new Response();
                // Write code to replicate the data to other nics
                // ReplicationHandler.replicateMetadata(packet);
                return replicationResponse;
            }); 
            Response response = futureReplication.get();     
        }
    }

    public void update(Packet packet) throws InterruptedException, ExecutionException,RocksDBException{
        Stopwatch stopWatch = Stopwatch.createUnstarted();
        ReplicationHandler replicationHandler = new ReplicationHandler();
        stopWatch.start();
        if (isMaster) {
            Response response = new Response();
            Future<Response> futureReplication = this.executorService.submit(()->{
                // Write code to replicate the data to other nics
                Response replicationResponse = new Response();
                // Write code to replicate the data to other nics
                replicationHandler.replicateMetadata(packet);
                return replicationResponse;
            });
        stopWatch.stop();
        logger.debug("Future time: " + stopWatch.elapsed(TimeUnit.MICROSECONDS));
        stopWatch.reset();
            // Future<Boolean> futureInvalidation = this.executorService.submit(()->{
            //    sendInvalidationRequest(packet);
            //     return false;
            // });
            
        stopWatch.start();
        response = futureReplication.get();
        stopWatch.stop();
        logger.debug("Replication time: " + stopWatch.elapsed(TimeUnit.MICROSECONDS));
        stopWatch.reset();
        // stopWatch.start();     
        // futureInvalidation.get();
        // stopWatch.stop();
        // logger.debug("Invalidation time: " + stopWatch.elapsed(TimeUnit.MICROSECONDS));
        logger.debug("Replicating complete Ack "+ response.getAck());
        }
        
    }

    // To implement delete
    public void delete(Packet packet) throws RocksDBException,InterruptedException,ExecutionException
    {
        ReplicationHandler replicationHandler = new ReplicationHandler();
        if (this.isMaster) {
            logger.debug("Starting replication");
            Future<Response> futureReplication = this.executorService.submit(()->{
                Response replicationResponse = new Response();
                // Write code to replicate the data to other nics
                replicationHandler.replicateMetadata(packet);
                return replicationResponse;
            }); 
            Response response = futureReplication.get();     
            logger.debug("Replicating complete "+ response.getAck());
        }
    }
    
//Paras method
   /* public ArrayList<DaRPCFuture<InvalidationRequest, InvalidationResponse>> sendInvalidateRequest(Request request){
        InvalidationRequest invalidationRequest = new InvalidationRequest();
        invalidationRequest.setKey(request.getKey());
        ArrayList<DaRPCFuture<InvalidationRequest,InvalidationResponse>> requestFutures;
        requestFutures = new ArrayList<>();
        for(var ep : invalidationClientEpStreams.keySet())
        {
            try
            {
                requestFutures.add(invalidationClientEpStreams.get(ep).
                                    request(invalidationRequest, new InvalidationResponse(), false));
            }catch(Exception e)
            {
                e.printStackTrace();
            }

        }
        return requestFutures;
    }/*
        // Pramod method
        /*
    public void sendInvalidateRequest(Request request) throws InterruptedException, ExecutionException{
        ExecutorService executorService = Executors.newWorkStealingPool();
        logger.debug("Sending Invalidation Request");
        InvalidationRequest invalidationRequest = new InvalidationRequest();
        invalidationRequest.setKey(request.getKey());
        Set<Callable<DaRPCFuture<InvalidationRequest, InvalidationResponse>>> callables = new HashSet<>();
        ArrayList<DaRPCFuture<InvalidationRequest,InvalidationResponse>> requestFutures = new ArrayList<>();;
        for (DaRPCStream<InvalidationRequest, InvalidationResponse> invalidationStream : NetworkHandler.invalidationStreams.values()) {
            callables.add(()->{
                InvalidationResponse response = new InvalidationResponse();
                DaRPCFuture<InvalidationRequest, InvalidationResponse> future = invalidationStream.request(invalidationRequest, response, false);
                return future;
            });
        }
        List<Future<DaRPCFuture<InvalidationRequest, InvalidationResponse>>> futures = executorService.invokeAll(callables);

        for (Future<DaRPCFuture<InvalidationRequest, InvalidationResponse>> future: futures) {
             InvalidationResponse response = future.get().get(); 
        }
        executorService.shutdown();
        
    }
    */
    public boolean consumeReplicationRequestFutures(ArrayList<DaRPCFuture<InvalidationRequest,InvalidationResponse>> requestFutures )
          throws InterruptedException,ExecutionException{

        //poll once 
        var it = requestFutures.iterator();
        while(it.hasNext())
        {
            var future = it.next();
            if(future.isDone())
            {
                it.remove();
            }
        }
        if(requestFutures.size() == 0)
            return true;

        //if not done now wait for 100 milliseconds before sending response
        requestFutures.get(0).get(100,TimeUnit.MILLISECONDS);
        
        it = requestFutures.iterator();
        while(it.hasNext())
        {
            var future = it.next();
            if(future.isDone())
            {
                it.remove();
            }
        }
        if(requestFutures.size() == 0)
            return true;
        //cancel each future
        requestFutures.forEach(future -> future.cancel(true));
        return false;

    }
}
