package hpdos_rdma_offloaded.handler;

import java.io.IOException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.ConcurrentLinkedQueue;

import com.ibm.disni.util.NativeAffinity;

import hpdos_rdma_offloaded.protocol.InvalidationRequest;
import hpdos_rdma_offloaded.protocol.InvalidationResponse;
import hpdos_rdma_offloaded.protocol.Request;
import hpdos_rdma_offloaded.protocol.Response;
import hpdos_rdma_offloaded.rdma.DaRPCFutureM;
import hpdos_rdma_offloaded.rdma.DaRPCStreamM;

public class NetworkHandlerM implements Runnable{

    private ConcurrentLinkedQueue<DaRPCStreamM<InvalidationRequest, InvalidationResponse>> invalidationStreams;
    private ConcurrentLinkedQueue<DaRPCStreamM<Request, Response>> replicationStreams;
    final static Logger logger = LoggerFactory.getLogger("NetworkHandler");

    public NetworkHandlerM()
    {
        this.invalidationStreams = new ConcurrentLinkedQueue<>();
        this.replicationStreams = new ConcurrentLinkedQueue<>();
    }
    public void addInvalidationStream(DaRPCStreamM<InvalidationRequest, InvalidationResponse> stream){
        this.invalidationStreams.add(stream);
    }

    public void addReplicationStream(DaRPCStreamM<Request,Response> stream)
    {
        logger.debug("size"+replicationStreams.size());
        this.replicationStreams.add(stream);
        logger.debug("size"+replicationStreams.size());
    }

    void processCloseInvalidationStream(DaRPCStreamM<InvalidationRequest,InvalidationResponse> stream)
    {
        stream.completedList.forEach(future -> 
        {
            if(future.metadataResponse.invalidationTotal == future.metadataResponse.invalidationDone.incrementAndGet())
                if(Response.DONE == future.metadataResponse.state.incrementAndGet())
                {
                    try
                    {
                        future.metadataResponse.event.triggerResponse();
                    }catch(IOException ex)
                    {
                        logger.error( ex.toString());   
                    }
                }
        });
        stream.endpoint.pendingFutures.forEach((ticket,future)->{
            if(future.metadataResponse.invalidationTotal == future.metadataResponse.invalidationDone.incrementAndGet())
                if(Response.DONE == future.metadataResponse.state.incrementAndGet())
                {
                    try
                    {
                        future.metadataResponse.event.triggerResponse();
                    }catch(IOException ex)
                    {
                        logger.error( ex.toString());   
                    }
                }
        });
    }
    void processInvalidationStream(DaRPCStreamM<InvalidationRequest,InvalidationResponse> stream)
    {
        try
        {
            DaRPCFutureM<InvalidationRequest,InvalidationResponse> future = stream.processStream();
            if(future!=null)
            {
                //logger.debug("future"+future.metadataResponse.invalidationTotal+" "+future.metadataResponse.invalidationDone);
                //logger.debug(future.metadataResponse.state);
                if(future.metadataResponse.invalidationTotal == future.metadataResponse.invalidationDone.incrementAndGet())
                    if(Response.DONE == future.metadataResponse.state.incrementAndGet())
                      future.metadataResponse.event.triggerResponse();           
            }
        }catch(Exception exx)
        {
            exx.printStackTrace();
        }
    }
    void processCloseReplicationStream(DaRPCStreamM<Request,Response> stream)
    {
        logger.info("stream closed");
        stream.completedList.forEach(future -> 
        {
            if(future.metadataResponse.replicationDone.incrementAndGet() == future.metadataResponse.replicatonTotal)
                if(Response.DONE == future.metadataResponse.state.incrementAndGet())                
                {
                    try
                    {
                        future.metadataResponse.event.triggerResponse();
                    }catch(IOException ex)
                    {
                        logger.error( ex.toString());   
                    }
                }
        });
        stream.endpoint.pendingFutures.forEach((ticket,future)->{
            if(future.metadataResponse.replicationDone.incrementAndGet() == future.metadataResponse.replicatonTotal)
                if(Response.DONE == future.metadataResponse.state.incrementAndGet())
                {
                    try
                    {
                        future.metadataResponse.event.triggerResponse();
                    }catch(IOException ex)
                    {
                        logger.error( ex.toString());   
                    }
                }
        });
    }
    void processReplicationStream(DaRPCStreamM<Request,Response> stream)
    {
        try
        {
            DaRPCFutureM<Request,Response> future = stream.processStream();
            if(future!=null)
            {
                if(future.metadataResponse.replicationDone.incrementAndGet() == future.metadataResponse.replicatonTotal)
                    if(Response.DONE == future.metadataResponse.state.incrementAndGet())
                      future.metadataResponse.event.triggerResponse();           
            }
        }catch(Exception exx)
        {
            exx.printStackTrace();
        }
    }
    public void run()
    {
        NativeAffinity.setAffinity(1<<1);
        while(true)
        {
            invalidationStreams.forEach(stream  ->
            {
                if( stream.endpoint.isClosed())
                {
                    processCloseInvalidationStream(stream);
                    invalidationStreams.remove(stream);
                }
                else
                {
                    processInvalidationStream(stream);
                }
            });
            replicationStreams.forEach(stream  ->
            {
                if(stream.endpoint.isClosed())
                {
                    processCloseReplicationStream(stream);
                    replicationStreams.remove(stream);
                }
                else
                {
                    processReplicationStream(stream);
                }
            });
        }
    }
    public void sendInvalidations(Request request,Response response)
    {
        InvalidationRequest inRequest =  new InvalidationRequest();
        //logger.debug("invdlidation send "+invalidationStreams.size());
        inRequest.setKey( request.getKey());

        int i = 0;
        for(var stream : invalidationStreams)
        {
            try
            {
                stream.request(inRequest, new InvalidationResponse(), response, true);
                ++i;
            }
            catch(IOException ex)
            {
                logger.error(ex.toString());
            }
        }   
        response.invalidationTotal = i;
        if(i == 0)
        {
            logger.debug("Zero invalidation");
            if(response.state.incrementAndGet() == Response.DONE)
            {
                try
                {
                    response.event.triggerResponse();
                }catch(IOException e)
                {
                    logger.error(e.getMessage());
                }
            }
        }
    }


    public void sendReplications(Request request,Response response)
    {
        /*Request repRequest =  new Request();
        repRequest.setRequestType(request.getRequestType());
        repRequest.setKey(request.getKey());

        if(request.getRequestType() == RequestType.PUT)
        {
            repRequest.setValue(request.getValue());
        }
        */
        int i = 0;
        logger.debug("sending Replication "+replicationStreams.size());
        for(var stream : replicationStreams)
        {
            try
            {
                stream.request(request, new Response(), response, true);
                ++i;
            }
            catch(IOException ex)
            {
                logger.error(ex.toString());
            }
        }  
        response.replicatonTotal = i;
        //This is because when there are no followers
        if(i == 0)
        {
            logger.debug("Zero replication");
            //This will happens only when there are no followers
            //and when this thread get delayed and invalidations got succesuful 
            if(response.state.incrementAndGet() == Response.DONE)
            {
                logger.debug("sending Response");
                try{
                    response.event.triggerResponse();
                }catch(IOException ex)
                {
                    logger.error(ex.getMessage());
                }
            }
        } 
    }
}
