package hpdos_rdma_offloaded.protocol;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.concurrent.atomic.AtomicInteger;

import com.ibm.darpc.DaRPCMessage;
import com.ibm.darpc.DaRPCServerEvent;

public class Response implements DaRPCMessage{

    public static int SERIALIZED_SIZE = 260;    
    public static int INVALIDATION_DONE = 1;
    public static int REPLICATION_DONE = 2;
    public static int DONE  = 3;
    public static int NEW = 0;

    public AtomicInteger invalidationDone;
    public AtomicInteger replicationDone;

    public int invalidationTotal;
    public int replicatonTotal ;
    public AtomicInteger state;

    public DaRPCServerEvent<Request,Response> event;


    private int ack;
    public byte[] key;
    public byte[] value; 

   

    public Response()
    {
        state = new AtomicInteger(NEW);
        replicationDone = new AtomicInteger(0);
        invalidationDone = new AtomicInteger(0);

    }
    public void setEvent(DaRPCServerEvent<Request,Response> event)
    {
        this.event = event;
    }

    public void triggerResponse() throws IOException{
        event.triggerResponse();
    }
    @Override
    public int write(ByteBuffer buffer) throws IOException 
    {
        buffer.putInt(ack);
        if(ack == AckType.SUCCESS_WITH_VALUE)
        {
            buffer.put(value);
            return 4+value.length;
        }
        return 4;
    }

    @Override
    public void update(ByteBuffer buffer) throws IOException
    {
        ack = buffer.getInt();
        if(ack == AckType.SUCCESS_WITH_VALUE)
        {
            if(value == null || value.length <128)
                value = new byte[128];
            buffer.get(value);
        }
    }

    @Override
    public int size() {
        
        return SERIALIZED_SIZE;
    }
    
    public int getAck(){
        return ack;
    }

    public void setAck(int ack){
        this.ack = ack;
    }

    public byte[] getKey() {
        return key;
    }

    public void setKey(byte[] key) {
        this.key = key;
    }

    public byte[] getValue() {
        return value;
    }

    public void setValue(byte[] value) {
        this.value = value;
    }
}
