package hpdos_rdma_offloaded.service;

import java.io.IOException;

import com.ibm.darpc.DaRPCClientGroup;
import com.ibm.darpc.DaRPCServerEndpoint;
import com.ibm.darpc.DaRPCServerEvent;
import com.ibm.darpc.DaRPCService;


import org.rocksdb.RocksDBException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import hpdos_rdma_offloaded.handler.StorageHandler;
import hpdos_rdma_offloaded.protocol.AckType;
import hpdos_rdma_offloaded.protocol.InvalidationRequest;
import hpdos_rdma_offloaded.protocol.InvalidationResponse;
import hpdos_rdma_offloaded.protocol.Request;
import hpdos_rdma_offloaded.protocol.RequestType;
import hpdos_rdma_offloaded.protocol.Response;
import hpdos_rdma_offloaded.protocol.RpcProtocol;

public class FollowerMetadataService extends RpcProtocol implements DaRPCService<Request, Response>{
    StorageHandler storageHandler;
    final static Logger logger = LoggerFactory.getLogger("Metadata service");
    DaRPCClientGroup<InvalidationRequest, InvalidationResponse> invalidationClientGroup;
    public FollowerMetadataService(StorageHandler storageHandler) throws Exception{
        logger.debug("New MetadataServerService Created");
        this.storageHandler  = storageHandler;
    }

    @Override
    public void processServerEvent(DaRPCServerEvent<Request, Response> event) throws IOException {
        Request request = event.getReceiveMessage();
        Response response = event.getSendMessage();
        
        logger.debug("Received "+request.getRequestType()+" Request for key: " + new String(request.getKey()));
        if(request.getValue() != null)
            logger.debug(" value : "+ new String(request.getValue()) );
        
        try{
            if (request.getRequestType() == RequestType.PUT) {
                this.storageHandler.create(request);
                response.setAck(AckType.SUCCESS);
            } 
            else if(request.getRequestType() == RequestType.GET) {
                byte[] value = this.storageHandler.read(request); 
                if(value == null)
                {
                    response.setAck(AckType.NOTFOUND);
                }
                else
                {
                    response.setAck(AckType.SUCCESS_WITH_VALUE);
                    response.setValue(value);
                }
            }
            else if(request.getRequestType() == RequestType.DELETE){
                this.storageHandler.delete(request);
                response.setAck(AckType.SUCCESS);
            }
            else{
                response.setAck(AckType.INVALID_REQUEST);
            }
        }
        catch(RocksDBException ex){
                ex.printStackTrace();
                response.setAck(AckType.DBFAILED);
        }
        event.triggerResponse();
    }

    @Override
    public void open(DaRPCServerEndpoint<Request, Response> rpcClientEndpoint)
    {
        try
        {
            logger.debug("Received new connection  "+rpcClientEndpoint.getDstAddr());
        }catch(Exception ex)
        {
            logger.error(ex.getMessage());
        }
    }

    @Override
    public void close(DaRPCServerEndpoint<Request, Response> rpcClientEndpoint)
    {
        logger.debug("Closing Connection");   
    }
    
}
